#include "stdint.h"
#include "s3c44b0x.h"
#include "delay.h"
#include "bitbang.h"
#include "uart.h"

void sck_high()
{
   *PDATF|=(1 << 8);
}

void sck_low()
{
   *PDATF&=~(1 << 8);
}

void mosi_high()
{
   *PDATF|=(1 << 5);
}

void mosi_low()
{
   *PDATF&=~(1 << 5);
}

void cs_high()
{
   *PDATC|=(1 << 11);
}

void cs_low()
{
   *PDATC&=~(1 << 11);
}


uint32_t read_bit()
{
   /* read state of MISO pin */
   return (((*PDATF) >> 7) & 0x01);
}

uint8_t mmc_byte(uint8_t x)
{
   uint32_t r=0;
   uint32_t i=0;
   for (i=0; i<8; i++)
   {
      sck_low();
      r=(r<<1);
      if (x & 0x80)
         mosi_high();
      else
         mosi_low();

      r|=read_bit();
      sck_high();
      x=(x<<1);
   }
   return r;
}

void mmc_command(uint8_t command, uint32_t arg)
{
   mmc_byte(0xff);
   mmc_byte(0x40 | command);
   mmc_byte((arg & 0xff000000) >> 24);
   mmc_byte((arg & 0xff0000) >> 16);
   mmc_byte((arg & 0xff00) >> 8);
   mmc_byte((arg & 0xff) >> 0);
   mmc_byte(0x95);
}

uint8_t mmc_reply(uint32_t byte_count)
{
   while (byte_count>1)
   {
      mmc_byte(0xff);
      --byte_count;
   }
   return mmc_byte(0xff);
}

void mmc_init()
{
   /* set pin registers to bit-bang an SPI interface
      chip select output on PC11
   */
   *PDATC|=(1 << 11);      /* preload SPI /CS to be high */
   *PCONC&=~(0x03 << 22);  /* mask out 2 bits for pin control */
   *PCONC|=(0x01 << 22);   /* 1=output */

   /* MOSI on PF5 */
   *PDATF|=(1 << 5);       /* preload MOSI to be high */
   *PCONF&=~(0x07 << 10);  /* mask out 3 bits for pin control */
   *PCONF|=(0x01 << 10);   /* 1=output (3=spi) */

   /* MISO on PF7 */
   *PUPF&=~(0x01 << 7);    /* 0=pullup on PF7 */
   *PCONF&=~(0x07 << 16);  /* mask out 3 bits for pin control */
   *PCONF|=(0x00 << 16);   /* 0=input (3=spi) */

   /* SCK on PF8 */
   *PDATF|=(1 << 8);       /* preload SCK high */
   *PCONF&=~(0x07 << 19);  /* mask out 3 bits for pin control */
   *PCONF|=(0x01 << 19);   /* 1=output (3=spi) */
}

void mmc_card_init()
{
   uint32_t i;

   mmc_init();
   cs_high();
   mosi_high();
   sck_high();

   /* reset the MMC state machine */
   for (i=0; i<12; i++)
      mmc_byte(0xff);

   cs_low();
   mmc_command(0x0, 0);
   delay_ms(1);
   mmc_reply(2);

   do
   {
      mmc_command(0x01, 0);
      delay_ms(1);
      i=mmc_reply(2);
      delay_ms(1);
   } while (i!=0);
}

void mmc_card_demo()
{
   uint32_t i;

   mmc_init();
   cs_high();
   mosi_high();
   sck_high();

   /* reset the MMC state machine */
   for (i=0; i<12; i++)
      mmc_byte(0xff);

   cs_low();
   mmc_command(0x0, 0);
   delay_ms(10);
   uart_tx_string("\r\nReset reply: 0x");
   uart_tx_string(itoh(mmc_reply(2)));
   delay_ms(10);

   mmc_command(58, 0);
   mmc_reply(3);
   for (i=0; i<4; i++)
   {
      uart_tx_string("\r\nREAD_OCR:0x");
      uart_tx_string(itoh(mmc_reply(1)));
   }
   delay_ms(20);

   do
   {
      mmc_command(0x01, 0);
      delay_ms(10);
      uart_tx_string("\r\nSPI mode switch reply: 0x");
      i=mmc_reply(2);
      uart_tx_string(itoh(i));
      delay_ms(10);
   } while (i!=0);

   mmc_dump_cid();

   uart_tx_string("\r\nAttempting to read MMC sector 0:\r\n");

   /* read sector command */
   mmc_command(17, 0);

   /* spin until we get a zero reply, indicating command accepted */
   do
   {
      i=mmc_reply(1);
   } while (i!=0);

   /* spin until start of data token is received (0xfe for single read) */
   do
   {
      i=mmc_reply(1);
   } while (i!=0xfe);

   /* Read from the card */
   for (i=0; i<128; i++)
   {
      if (i%8)
         uart_tx_string("  ");
      else
      uart_tx_string("\r\n");

      uart_tx_string(itoh(((mmc_reply(1) << 0) |
                           (mmc_reply(1) << 8) |
                           (mmc_reply(1) << 16) |
                           (mmc_reply(1) << 24))));

   }

   /* give the MMC state machine 8 extra clocks */
   mmc_byte(0xff);
}

void mmc_write_sector(uint32_t sector, uint8_t* start)
{
   uint32_t i;

   /* let's write a sector */
   mmc_command(24, sector);
   uart_tx_string("\r\nSector 0x");
   uart_tx_string(itoh(sector/0x200));
   uart_tx_string(": reply: 0x");
   uart_tx_string(itoh(mmc_reply(2)));
   uart_tx_string("...writing...");

   mmc_byte(0xfe);
   for (i=0; i<512; i++)
   {
      mmc_byte(*start);
      start++;
   }

   /* 2 crc bytes (ignored) */
   mmc_byte(0);
   mmc_byte(0);
   i=mmc_reply(1);
   if (0x05 == (i & 0x0f))
      uart_tx_string("ok.");
   else
   {
      uart_tx_string("error: 0x");
      uart_tx_string(itoh(i & 0x0f));
   }

   /* spin until it's not busy anymore */
   do
   {
      i=mmc_byte(0xff);
   } while (0==i);
}

void mmc_dump_cid()
{
   uint32_t i;

   /* read CID register (16 bytes) */
   mmc_command(10, 0);
   uart_tx_string("\r\nRead Card ID register");
   mmc_reply(4);     /* skip 3 bytes of reply */

   i=mmc_reply(1);
   uart_tx_string("\r\nManufacturer ID: 0x");
   uart_tx_string(itoh(i));

   i=mmc_reply(1);
   i=(i << 8) | mmc_reply(1);
   uart_tx_string("\r\nOEM/App ID: 0x");
   uart_tx_string(itoh(i));

   uart_tx_string("\r\nProduct name: ");
   for (i=0; i<6; i++)
   {
      uart_tx_char(mmc_reply(1));
   }

   uart_tx_string("\r\nProduct revision: 0x");
   uart_tx_string(itoh(mmc_reply(1)));

   uart_tx_string("\r\nSerial number: 0x");
   i=mmc_reply(1);
   i=(i << 8) | mmc_reply(1);
   i=(i << 8) | mmc_reply(1);
   i=(i << 8) | mmc_reply(1);
   uart_tx_string(itoh(i));

   uart_tx_string("\r\nManufacturing date code: 0x");
   uart_tx_string(itoh(mmc_reply(1)));

   uart_tx_string("\r\nCRC: 0x");
   uart_tx_string(itoh(mmc_reply(1)));

}

void mmc_idle()
{
   cs_high();
   mmc_byte(0xff);
   mmc_byte(0xff);
}
